from django.db import models

from . import InvoiceLine


class SerialNumber(models.Model):
    line = models.ForeignKey(
        InvoiceLine, on_delete=models.CASCADE, related_name='serials',
        verbose_name='fakturalinje'
    )
    serial = models.CharField(primary_key=True, max_length=50,
                              verbose_name='serienummer')

    class Meta:
        unique_together = ['line', 'serial']
        verbose_name = 'serienummer'
        verbose_name_plural = 'serienummer'

    def __str__(self):
        return self.serial



from django.contrib import admin

from ehf_invoice.models import Invoice, InvoiceLine, SerialNumber, \
    Attachment, Customer, Supplier


@admin.register(Invoice)
class InvoiceAdmin(admin.ModelAdmin):
    list_filter = ['supplier']
    list_display = ['supplier', 'invoice_number', 'date']
    readonly_fields = ['invoice_lines']


@admin.register(InvoiceLine)
class InvoiceLineAdmin(admin.ModelAdmin):
    list_display = ['invoice_number', 'line_id', 'name']
    readonly_fields = ['invoice']

    def invoice_number(self, obj: InvoiceLine):
        return obj.invoice_number()
    invoice_number.short_description = 'Fakturanummer'


@admin.register(SerialNumber)
class SerialNumberAdmin(admin.ModelAdmin):
    list_display = ['line', 'serial']
    list_filter = ['line__invoice__supplier']
    readonly_fields = ['line', 'serial']


@admin.register(Attachment)
class AttachmentAdmin(admin.ModelAdmin):
    list_display = ['invoice', 'file']


@admin.register(Customer)
class CustomerAdmin(admin.ModelAdmin):
    list_display = ['id', 'name']
    readonly_fields = list_display


@admin.register(Supplier)
class SupplierAdmin(admin.ModelAdmin):
    list_display = ['id', 'name']
    readonly_fields = list_display



from decimal import Decimal


class InvoiceLine:
    line = None
    namespaces = {
        'cac': 'urn:oasis:names:specification:ubl:'
        'schema:xsd:CommonAggregateComponents-2',
        'cbc': 'urn:oasis:names:specification:ubl:'
        'schema:xsd:CommonBasicComponents-2',
    }

    def __init__(self, line):
        self.line = line

    @property
    def id(self):
        return self.line.find('cbc:ID', self.namespaces).text

    @property
    def quantity(self):
        q = self.line.find('cbc:InvoicedQuantity', self.namespaces).text
        # q = q[:-3]
        return int(float(q))

    @property
    def description(self):
        desc = self.line.find('cac:Item/cbc:Description', self.namespaces)
        if desc:
            return desc.text

    @property
    def name(self):
        return self.line.find('cac:Item/cbc:Name', self.namespaces).text

    @property
    def serials(self):
        properties = self.line.findall(
            'cac:Item/cac:AdditionalItemProperty/'
            'cbc:Name[.="SerialNumber"]/../cbc:Value',
            self.namespaces,
        )
        serials = []
        for serial in properties:
            serials.append(serial.text)
        return serials

    @property
    def price(self):
        """
        Price of one item
        :return:
        """
        return Decimal(
            self.line.find('cac:Price/cbc:PriceAmount', self.namespaces).text
        )

    @property
    def sum(self):
        """
        Sum of the line
        :return:
        """
        return Decimal(
            self.line.find('cbc:LineExtensionAmount', self.namespaces).text
        )



from django.contrib.auth.decorators import permission_required
from django.shortcuts import render

from ehf_invoice.models import Invoice, SerialNumber


@permission_required('ehf_invoice.view_invoice', raise_exception=True)
def show_invoice(request, invoice_number):
    invoice = Invoice.objects.get(invoice_number=invoice_number)
    return render(request, 'ehf_invoice/invoice.html', {'invoice': invoice})


@permission_required('ehf_invoice.view_serialnumber', raise_exception=True)
def find_serial(request, serial_number):
    serial = SerialNumber.objects.get(serial=serial_number)
    invoice = serial.line.invoice
    return render(request, 'ehf_invoice/invoice.html', {'invoice': invoice})



from django.urls import path

from ehf_invoice import views

app_name = 'ehf_invoice'
urlpatterns = [
    path(
        'invoice/<str:invoice_number>', views.show_invoice, name='show_invoice'
    ),
    path(
        'serial/<str:serial_number>',
        views.find_serial,
        name='find_serial_number',
    ),
]



from base64 import b64decode
from decimal import Decimal
from xml.etree import ElementTree

from ehf_invoice.parser import Attachment
from ehf_invoice.parser.InvoiceLine import InvoiceLine
from ehf_invoice.parser.Party import Party


class InvoiceXML:
    namespaces = {
        'cac': 'urn:oasis:names:specification:'
               'ubl:schema:xsd:CommonAggregateComponents-2',
        'cbc': 'urn:oasis:names:specification:'
               'ubl:schema:xsd:CommonBasicComponents-2',
    }

    invoice = None

    def __init__(self, file):
        # import xml.etree.ElementTree
        f = open(file, 'r')
        self.invoice = ElementTree.fromstring(f.read())
        f.close()

    def find(self, match):
        """

        :param str match:
        :return Element:
        """
        return self.invoice.find(match, self.namespaces)

    @property
    def invoice_number(self):
        return self.find('cbc:ID').text

    @property
    def invoice_date(self):
        return self.find('cbc:IssueDate').text

    @property
    def customer(self):
        customer = self.find('cac:AccountingCustomerParty/cac:Party')
        return Party(customer)

    @property
    def supplier(self):
        supplier = self.find('cac:AccountingSupplierParty/cac:Party')
        return Party(supplier)

    # @property
    # def supplier_name(self):
    #     return (
    #         self.supplier
    #             .find('cac:PartyName/cbc:Name', self.namespaces)
    #             .text
    #     )

    @property
    def order_reference(self):
        reference = self.find('cac:OrderReference/cbc:ID')
        if reference:
            return reference

    @property
    def order_number(self):
        objects = self.invoice.findall(
            'cac:ContractDocumentReference/cbc:ID', self.namespaces
        )
        references = []
        reference1 = self.order_reference
        for reference in objects:
            if not reference.text == reference1:
                return reference.text
            # references.append(reference.text)
        # return references

    def attachments(self):
        objects = self.invoice.findall('cac:AdditionalDocumentReference/cac:Attachment', self.namespaces)
        attachments = []
        for att in objects:
            try:
                attachments.append(Attachment(att))
            except ValueError:
                continue
        return attachments

    @property
    def amount(self):
        return Decimal(
            self.find('cac:LegalMonetaryTotal/cbc:TaxExclusiveAmount').text
        )

    def invoice_lines(self):
        lines_xml = self.invoice.findall('cac:InvoiceLine', self.namespaces)
        lines = []
        for line in lines_xml:
            lines.append(InvoiceLine(line))
        return lines



from django.db import models


class Supplier(models.Model):
    id = models.IntegerField('organisasjonsnummer', primary_key=True)
    name = models.CharField('navn', max_length=200)

    class Meta:
        verbose_name = 'leverandør'
        verbose_name_plural = 'leverandører'

    def __str__(self):
        return self.name



from base64 import b64decode


class Attachment:
    namespaces = {
        'cac': 'urn:oasis:names:specification:'
               'ubl:schema:xsd:CommonAggregateComponents-2',
        'cbc': 'urn:oasis:names:specification:'
               'ubl:schema:xsd:CommonBasicComponents-2',
    }

    def __init__(self, element):
        self.element = element

        file = element.find('cbc:EmbeddedDocumentBinaryObject',
                            self.namespaces)
        if file is None:
            raise ValueError('No file in attachment')
        if 'filename' not in file.attrib:
            raise ValueError('Attachment has no file name')

        self.file_name = file.attrib['filename']
        self.mime = file.attrib['mimeCode']

        if 'format' not in file.attrib or file.attrib['format'] == 'BASE64':
            self.data = b64decode(file.text, validate=True)
        else:
            raise ValueError(
                'Unknown attachment format: %s' % file.attrib['format'])



import os

from django.core.files.base import ContentFile
from django.core.management.base import BaseCommand

from ehf_invoice.models import (
    Invoice,
    InvoiceLine,
    SerialNumber,
    Attachment,
    Supplier,
    Customer,
)
from ehf_invoice.parser import InvoiceXML


class Command(BaseCommand):
    help = 'Load invoice'

    def add_arguments(self, parser):
        parser.add_argument('file', nargs='+', type=str)

    def load(self, file: str) -> Invoice:
        invoice = InvoiceXML(file)
        supplier_xml = invoice.supplier
        supplier, created = Supplier.objects.get_or_create(
            id=supplier_xml.id, defaults={'name': supplier_xml.name}
        )

        customer, created = Customer.objects.get_or_create(
            id=invoice.customer.id, defaults={'name': invoice.customer.name}
        )

        try:
            invoice_obj = Invoice.objects.get(
                supplier=supplier, invoice_number=invoice.invoice_number
            )
        except Invoice.DoesNotExist:
            invoice_obj = Invoice(
                supplier=supplier, invoice_number=invoice.invoice_number
            )

        invoice_obj.customer = customer
        invoice_obj.order_number = invoice.order_number
        invoice_obj.date = invoice.invoice_date
        invoice_obj.amount = invoice.amount

        invoice_obj.save()
        print(invoice_obj)

        for attachment in invoice.attachments():
            try:
                Attachment.objects.get(
                    invoice=invoice_obj, name=attachment.file_name
                )
                continue
            except Attachment.DoesNotExist:
                file = ContentFile(attachment.data)
                file.name = attachment.file_name
                attachment_obj = Attachment(
                    invoice=invoice_obj,
                    mime=attachment.mime,
                    file=file,
                    name=attachment.file_name,
                )
                attachment_obj.save()

        for line in invoice.invoice_lines():
            line_obj, created = InvoiceLine.objects.get_or_create(
                invoice=invoice_obj,
                line_id=line.id,
                defaults={
                    'quantity': line.quantity,
                    'description': line.description,
                    'name': line.name,
                    'price': line.price,
                    'sum': line.sum,
                },
            )

            for serial in line.serials:
                serial_obj = SerialNumber(line=line_obj, serial=serial)
                serial_obj.save()
        return invoice_obj

    def handle(self, *args, **options):
        if os.path.isfile(options['file'][0]):
            return self.load(options['file'][0])
        elif os.path.isdir(options['file'][0]):
            for file in os.scandir(options['file'][0]):
                self.load(file)



class Party:
    namespaces = {
        'cac': 'urn:oasis:names:specification:'
               'ubl:schema:xsd:CommonAggregateComponents-2',
        'cbc': 'urn:oasis:names:specification:'
               'ubl:schema:xsd:CommonBasicComponents-2',
    }

    def __init__(self, element):
        self.id = element.find('cbc:EndpointID', self.namespaces).text
        self.name = element.find('cac:PartyName/cbc:Name', self.namespaces).text



import os
import unittest

from ehf_invoice.parser import InvoiceXML, Party


class ParserTestCase(unittest.TestCase):
    def test_parser(self):
        invoice = InvoiceXML(
            os.path.join(os.path.dirname(__file__), 'test_data',
                         'T10 BII05 gyldig faktura.xml'))
        self.assertEqual(1436.5, invoice.amount)
        self.assertIsInstance(invoice.customer, Party)
        self.assertIsInstance(invoice.supplier, Party)
        self.assertEqual('2013-06-30', invoice.invoice_date)
        self.assertEqual('TOSL108', invoice.invoice_number)
        self.assertEqual('Contract321', invoice.order_number)


if __name__ == '__main__':
    unittest.main()



from django.db import models

from ehf_invoice.models import Supplier, Customer


class Invoice(models.Model):
    supplier = models.ForeignKey(
        Supplier, on_delete=models.PROTECT, verbose_name='leverandør'
    )
    customer = models.ForeignKey(
        Customer, on_delete=models.PROTECT, verbose_name='kunde'
    )
    invoice_number = models.CharField('fakturanummer', max_length=100)
    order_number = models.CharField('ordrenummer', max_length=100, blank=True, null=True)
    date = models.DateField('fakturadato')
    amount = models.DecimalField('beløp', decimal_places=2, max_digits=12)

    class Meta:
        unique_together = ['supplier', 'invoice_number']
        verbose_name = 'faktura'
        verbose_name_plural = 'fakturaer'
        ordering = ['-date']

    def __str__(self):
        return '%s invoice %s' % (self.supplier, self.invoice_number)

    def invoice_lines(self):
        return self.lines.all()



from django.db import models

from . import Invoice


class InvoiceLine(models.Model):
    invoice = models.ForeignKey(
        Invoice, on_delete=models.CASCADE, related_name='lines',
        verbose_name='faktura'
    )
    line_id = models.IntegerField('linje')

    description = models.CharField('beskrivelse', max_length=300, null=True,
                                   blank=True)
    name = models.CharField('navn', max_length=300)
    price = models.DecimalField('beløp', decimal_places=2, max_digits=12)
    quantity = models.IntegerField('antall')
    sum = models.DecimalField('sum', decimal_places=2, max_digits=12)

    class Meta:
        unique_together = ['invoice', 'line_id']
        verbose_name = 'fakturalinje'
        verbose_name_plural = 'fakturalinjer'
        ordering = ['invoice', 'line_id']

    def __str__(self):
        return '%s line %s %s' % (self.invoice, self.line_id, self.name)

    def invoice_number(self):
        return self.invoice.invoice_number



# flake8: noqa
from .Customer import Customer
from .Supplier import Supplier
from .Invoice import Invoice
from .InvoiceLine import InvoiceLine
from .SerialNumber import SerialNumber
from .Attachment import Attachment



from django.db import models
from django.db.models.signals import post_delete
from django.dispatch import receiver

from ehf_invoice.models import Invoice


class Attachment(models.Model):
    invoice = models.ForeignKey(
        Invoice, on_delete=models.CASCADE, related_name='attachments'
    )
    name = models.CharField(max_length=255)
    file = models.FileField(upload_to='invoice_attachments/%Y/')
    mime = models.CharField('MIME type', max_length=50)

    class Meta:
        unique_together = ['invoice', 'name']
        verbose_name = 'vedlegg'
        verbose_name_plural = 'vedlegg'


@receiver(post_delete, sender=Attachment)
def attachment_delete(sender, instance: Attachment, **kwargs):
    """
    Receive the post_delete signal
    and delete the file associated with the model instance.
    https://stackoverflow.com/a/14310174/2630074
    """
    # Pass false so FileField doesn't save the model.
    instance.file.delete(False)



# Generated by Django 3.1.5 on 2021-02-11 20:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Customer',
            fields=[
                ('id', models.IntegerField(primary_key=True, serialize=False, verbose_name='organisasjonsnummer')),
                ('name', models.CharField(max_length=200, verbose_name='navn')),
            ],
            options={
                'verbose_name': 'kunde',
                'verbose_name_plural': 'kunder',
            },
        ),
        migrations.CreateModel(
            name='Invoice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('invoice_number', models.CharField(max_length=100, verbose_name='fakturanummer')),
                ('order_number', models.CharField(blank=True, max_length=100, null=True, verbose_name='ordrenummer')),
                ('date', models.DateField(verbose_name='fakturadato')),
                ('amount', models.DecimalField(decimal_places=2, max_digits=12, verbose_name='beløp')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ehf_invoice.customer', verbose_name='kunde')),
            ],
            options={
                'verbose_name': 'faktura',
                'verbose_name_plural': 'fakturaer',
                'ordering': ['-date'],
            },
        ),
        migrations.CreateModel(
            name='InvoiceLine',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('line_id', models.IntegerField(verbose_name='linje')),
                ('description', models.CharField(blank=True, max_length=300, null=True, verbose_name='beskrivelse')),
                ('name', models.CharField(max_length=300, verbose_name='navn')),
                ('price', models.DecimalField(decimal_places=2, max_digits=12, verbose_name='beløp')),
                ('quantity', models.IntegerField(verbose_name='antall')),
                ('sum', models.DecimalField(decimal_places=2, max_digits=12, verbose_name='sum')),
                ('invoice', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='lines', to='ehf_invoice.invoice', verbose_name='faktura')),
            ],
            options={
                'verbose_name': 'fakturalinje',
                'verbose_name_plural': 'fakturalinjer',
                'ordering': ['invoice', 'line_id'],
                'unique_together': {('invoice', 'line_id')},
            },
        ),
        migrations.CreateModel(
            name='Supplier',
            fields=[
                ('id', models.IntegerField(primary_key=True, serialize=False, verbose_name='organisasjonsnummer')),
                ('name', models.CharField(max_length=200, verbose_name='navn')),
            ],
            options={
                'verbose_name': 'leverandør',
                'verbose_name_plural': 'leverandører',
            },
        ),
        migrations.AddField(
            model_name='invoice',
            name='supplier',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ehf_invoice.supplier', verbose_name='leverandør'),
        ),
        migrations.CreateModel(
            name='SerialNumber',
            fields=[
                ('serial', models.CharField(max_length=50, primary_key=True, serialize=False, verbose_name='serienummer')),
                ('line', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='serials', to='ehf_invoice.invoiceline', verbose_name='fakturalinje')),
            ],
            options={
                'verbose_name': 'serienummer',
                'verbose_name_plural': 'serienummer',
                'unique_together': {('line', 'serial')},
            },
        ),
        migrations.AlterUniqueTogether(
            name='invoice',
            unique_together={('supplier', 'invoice_number')},
        ),
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('file', models.FileField(upload_to='invoice_attachments/%Y/')),
                ('mime', models.CharField(max_length=50, verbose_name='MIME type')),
                ('invoice', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='ehf_invoice.invoice')),
            ],
            options={
                'verbose_name': 'vedlegg',
                'verbose_name_plural': 'vedlegg',
                'unique_together': {('invoice', 'name')},
            },
        ),
    ]



import os
import datetime

from django.test import TestCase
from ehf_invoice.management.commands.load_invoice import Command
from ehf_invoice.models import Invoice


class ImportTestCase(TestCase):
    def testImport(self):
        file = os.path.join(os.path.dirname(__file__), 'test_data',
                            'T10 BII05 gyldig faktura.xml')
        load = Command()
        load.load(file)
        invoice = Invoice.objects.get(invoice_number='TOSL108')
        self.assertIsInstance(invoice, Invoice)
        self.assertEqual(datetime.date(2013, 6, 30), invoice.date)
        self.assertEqual('TOSL108', invoice.invoice_number)
        self.assertEqual('Contract321', invoice.order_number)
